@@ -1,13 +1,20 @@
 # functions.py
 """PyCMTensor functions module"""
+from typing import Union
+
 import aesara.tensor as aet
 import aesara.tensor.nlinalg as nlinalg
 import numpy as np
+from aesara.tensor.var import TensorVariable
 
-from .logger import log
+from pycmtensor.expressions import Beta
+
+from .logger import error, log
 
 
-def exp_mov_average(batch_avg, moving_avg, alpha=0.1):
+def exp_mov_average(
+    batch_avg: TensorVariable, moving_avg: TensorVariable, alpha: float = 0.1
+):
     """Calculates the exponential moving average (EMA) of a new minibatch
 
     Args:
@@ -37,13 +44,17 @@ def exp_mov_average(batch_avg, moving_avg, alpha=0.1):
     return ema
 
 
-def logit(utility: list, avail: list = None):
+def logit(
+    utility: Union[list, tuple, TensorVariable],
+    avail: Union[list, tuple, TensorVariable] = None,
+):
     """Computes the Logit function, with availability conditions.
 
     Args:
-        utility (list): list of :math:`M` utility equations
-        avail (list): list of :math:`M` availability conditions, if no availability
-            conditions are provided, defaults to 1 for all availabilities.
+        utility (list, tuple, TensorVariable): list of :math:`M` utility equations
+        avail (list, tuple, TensorVariable): list of :math:`M` availability conditions,
+            if no availability conditions are provided, defaults to 1 for all
+            availabilities.
 
     Returns:
         TensorVariable: A NxM matrix of probabilities.
@@ -60,7 +71,7 @@ def logit(utility: list, avail: list = None):
     else:
         U = utility
 
-    prob = aet.nnet.softmax(U, axis=0)
+    prob = aet.special.softmax(U, axis=0)
     if avail != None:
         AV = aet.stack(avail)
         while AV.ndim < prob.ndim:
@@ -70,7 +81,7 @@ def logit(utility: list, avail: list = None):
     return prob
 
 
-def log_likelihood(prob, y):
+def log_likelihood(prob: TensorVariable, y: TensorVariable):
     """Symbolic representation of the log likelihood cost function.
 
     Args:
@@ -86,18 +97,19 @@ def log_likelihood(prob, y):
     return aet.sum(aet.log(prob)[y, ..., aet.arange(y.shape[0])])
 
 
-def rmse(y_est, y):
-    """Computes the root mean squared error between pairs of observations
+def rmse(y_hat: TensorVariable, y: TensorVariable):
+    """Computes the root mean squared error (RMSE) between pairs of observations
 
     Args:
-        y_est (TensorVariable): model estimated values, vector
-        y (TensorVariable): ground truth values, vector
+        y_hat (TensorVariable): model estimated values
+        y (TensorVariable): ground truth values
 
     Returns:
         TensorVariable: symbolic scalar representation of the rmse
 
     Note:
-        Tensor is flattened to an N-vector if the input args are :math:`Nx1` matrices.
+        Tensor is flattened to an N-vector if the input args are :math:`N\\times 1`
+        matrices.
 
         Formula:
 
@@ -106,29 +118,28 @@ def rmse(y_est, y):
             RMSE = \\sqrt{\\frac{1}{N}\\sum_{i=1}^N(\\hat{y}_i-y_i)^2}
 
     """
-    if y_est.ndim != y.ndim:
-        msg = f"y_est should have the same dimensions as y. y_est.ndim: {y_est.ndim}, q.ndim: {y.ndim}"
-        log(40, msg)
+    if y_hat.ndim != y.ndim:
+        msg = f"y_hat should have the same dimensions as y. y_hat.ndim: {y_hat.ndim}, q.ndim: {y.ndim}"
         raise ValueError(msg)
-    if y_est.ndim < 1:
-        y_est = aet.flatten(y_est)
+    if y_hat.ndim < 1:
+        y_hat = aet.flatten(y_hat)
         y = aet.flatten(y)
 
-    return aet.sqrt(aet.mean(aet.sqr(y_est - y)))
+    return aet.sqrt(aet.mean(aet.sqr(y_hat - y)))
 
 
-def mae(y_est, y):
-    """Computes the mean absolute error between pairs of observations
+def mae(y_hat: TensorVariable, y: TensorVariable):
+    """Computes the mean absolute error (MAE) between pairs of observations
 
     Args:
-        y_est (TensorVariable): model estimated values, vector
-        y (TensorVariable): ground truth values, vector
+        y_hat (TensorVariable): model estimated values
+        y (TensorVariable): ground truth values
 
     Returns:
         TensorVariable: symbolic scalar representation of the mae
 
     Note:
-        Tensor is flattened to an N-vector if the input args are :math:`Nx1` matrices.
+        Tensor is flattened to an N-vector if the input args are :math:`N\\times 1` matrices.
 
         Formula:
 
@@ -136,18 +147,17 @@ def mae(y_est, y):
 
             MAE = \\frac{\sum_{i=1}^N|\\hat{y}_i-y_i|}{N}
     """
-    if y_est.ndim != y.ndim:
-        msg = f"y_est should have the same dimensions as y. y_est.ndim: {y_est.ndim}, q.ndim: {y.ndim}"
-        log(40, msg)
+    if y_hat.ndim != y.ndim:
+        msg = f"y_hat should have the same dimensions as y. y_hat.ndim: {y_hat.ndim}, q.ndim: {y.ndim}"
         raise ValueError(msg)
-    if y_est.ndim < 1:
-        y_est = aet.flatten(y_est)
+    if y_hat.ndim < 1:
+        y_hat = aet.flatten(y_hat)
         y = aet.flatten(y)
 
-    return aet.mean(aet.abs(y_est - y))
+    return aet.mean(aet.abs(y_hat - y))
 
 
-def kl_divergence(p, q):
+def kl_divergence(p: TensorVariable, q: TensorVariable):
     """Computes the KL divergence loss between discrete distributions ``p`` and ``q``.
 
     Args:
@@ -169,12 +179,12 @@ def kl_divergence(p, q):
     """
     if p.ndim != q.ndim:
         msg = f"p should have the same shape as q. p.ndim: {p.ndim}, q.ndim: {q.ndim}"
-        log(40, msg)
         raise ValueError(msg)
+
     return aet.sum(aet.switch(aet.neq(p, 0), p * aet.log(p / q), 0))
 
 
-def kl_multivar_norm(m0, v0, m1, v1):
+def kl_multivar_norm(m0, v0, m1, v1, epsilon=1e-6):
     """Computes the KL divergence loss between two multivariate normal distributions.
 
     Args:
@@ -182,11 +192,9 @@ def kl_multivar_norm(m0, v0, m1, v1):
         v0: (co-)variance matrix of the first Normal m.v. distribution :math:`N_0`
         m1: mean vector of the second Normal m.v. distribution :math:`N_1`
         v1: (co-)variance of the second Normal m.v. distribution :math:`N_1`
+        epsilon (float): small value to prevent divide-by-zero error
 
     Note:
-        If m0 and v0 are 0 and 1 respectively, returns a univariate norm solution.
-        If m1 and v1 are 0 and 1 respectively, computes a simplified version of the multivariate norm.
-
         k = dimension of the distribution.
 
         Formula:
@@ -194,6 +202,12 @@ def kl_multivar_norm(m0, v0, m1, v1):
         .. math::
 
             D_{KL}(N_0||N_1) = 0.5 * \\Big(\\ln\\big(\\frac{|v_1|}{|v_0|}\\big) + trace(v_1^{-1} v_0) + (m_1-m_0)^T v_1^{-1} (m_1-m_0) - k\\Big)
+
+        In variational inference, the kl divergence is the relative entropy between a
+        diagonal multivariate Normal and a standard Normal distribution, N(0, 1),
+        therefore, for VI, ``m1=aet.constant(0)``, ``v1=aet.constant(1)``
+
+        For two univariate distributions, dimensions of m0,m1,v0,v1 = 0
     """
     if not (
         (m0.ndim >= m1.ndim)
@@ -202,30 +216,39 @@ def kl_multivar_norm(m0, v0, m1, v1):
         and (v0.ndim <= 2)
     ):
         msg = f"Incorrect dimensions inputs: m0.ndim={m0.ndim}, v0.ndim={v0.ndim}, m1.ndim={m1.ndim}, v1.ndim={v1.ndim}"
-        log(40, msg)
         raise ValueError(msg)
 
-    if (m0.ndim == v0.ndim == 0) or (m1.ndim == v1.ndim == 0):
-        # computes univariate norm or multivariate norm with N(m1, v1)=N(0, 1)
-        if v0.ndim == 2:
-            v0 = aet.diag(v0)
-        return aet.sum(0.5 * ((v0 + aet.sqr(m0 - m1)) / v1 - aet.log(v0 / v1) - 1))
+    # VI KL divergence
+    if (m1.ndim == v1.ndim == 0) and (m0.ndim == 1) and (v0.ndim == 2):
+        v0 = aet.diag(v0)
+        kld = 0.5 * aet.sum(v0 + aet.sqr(m0) - 1 - aet.log(v0))
 
-    k = m0.shape[0]
-    v1_inv = nlinalg.inv(v1)
-    det_term = aet.log(nlinalg.det(v1) / nlinalg.det(v0))
-    trace_term = nlinalg.trace(aet.dot(v1_inv, v0))
-    return aet.sum(
-        det_term + trace_term + aet.dot((m1 - m0).T, aet.dot(v1_inv, (m1 - m0))) - k
-    )
+    # univariate KL divergence
+    elif (m0.ndim == v0.ndim == 0) and (m1.ndim == v1.ndim == 0):
+        kld = (
+            aet.log(aet.sqrt(v1 / (v0 + epsilon)))
+            + 0.5 * (v0 + aet.sqr(m0 - m1)) / v1
+            - 0.5
+        )
+
+    # multivariate KL divergence
+    else:
+        k = m0.shape[0]
+        v1_inv = nlinalg.inv(v1)
+        det_term = aet.log(nlinalg.det(v1) / nlinalg.det(v0))
+        trace_term = nlinalg.trace(aet.dot(v1_inv, v0))
+        prod_term = aet.dot((m1 - m0).T, aet.dot(v1_inv, (m1 - m0)))
+        kld = aet.sum(det_term + trace_term + prod_term - k)
+
+    return kld
 
 
-def errors(prob, y):
-    """Symbolic representation of the prediction as a percentage error.
+def errors(prob: TensorVariable, y: TensorVariable):
+    """Symbolic representation of the discrete prediction as a percentage error.
 
     Args:
         prob (TensorVariable): matrix describing the choice probabilites
-        y (TensorVariable): The ``TensorVariable`` referencing the choice column
+        y (TensorVariable): the ``TensorVariable`` referencing the choice column
 
     Returns:
         TensorVariable: the mean prediction error over the input ``y``
@@ -238,7 +261,7 @@ def errors(prob, y):
         raise NotImplementedError(f"y should be int32 or int64", ("y.dtype:", y.dtype))
 
 
-def hessians(ll, params):
+def hessians(ll: TensorVariable, params: list[Beta]):
     """Symbolic representation of the Hessian matrix given the log likelihood.
 
     Args:
@@ -246,17 +269,13 @@ def hessians(ll, params):
         params (list): list of params to compute the gradients over
 
     Returns:
-        TensorVariable: the Hessian matrix with ndim=2
+        TensorVariable: the Hessian matrix
 
     Note:
         Parameters with status=1 are ignored.
     """
-    if not isinstance(params, (dict, list)):
-        raise TypeError(
-            f"params is not list or dict instance. type(params)={type(params)}"
-        )
-    if isinstance(params, dict):
-        params = list(params.values())
+    if not isinstance(params, list):
+        raise TypeError(f"params is not list instance. type(params)={type(params)}")
     params = [p() for p in params if (p.status != 1)]
     grads = aet.grad(ll, params, disconnected_inputs="ignore")
     mat = aet.as_tensor_variable(np.zeros((len(grads), len(grads))))
@@ -268,7 +287,7 @@ def hessians(ll, params):
     return mat
 
 
-def bhhh(ll, params):
+def bhhh(ll: TensorVariable, params: list[Beta]):
     """Symbolic representation of the Berndt-Hall-Hall-Hausman (BHHH) algorithm given
     the log likelihood.
 
@@ -294,7 +313,7 @@ def bhhh(ll, params):
     return mat
 
 
-def gnorm(cost, params):
+def gnorm(cost: TensorVariable, params: list[Beta]):
     """Symbolic representation of the gradient norm given the cost.
 
     Args:
